from abc import abstractmethod
from typing import Callable, Generic, final, TypeVar

class StateCell:
  @abstractmethod
  def serlialize(self) -> str: pass
  def destroy(self): pass

@final
class StrStateCell(StateCell):
  def __init__(self, value: str) -> None:
    self.value = value

  def serlialize(self): return self.value

T = TypeVar("T")
class SerilializableStateCell(StateCell, Generic[T]):
  def __init__(self, value: T, serializer: Callable[[T], str]) -> None:
    super().__init__()
    self.value = value
    self._serializer = serializer

  def serlialize(self) -> str: return self._serializer(self.value)



import unittest
from rxxxt.component import Component
from rxxxt.elements import El, HTMLFragment, VEl
from rxxxt.utils import class_map
from tests.helpers import render_element

class TestElements(unittest.IsolatedAsyncioTestCase):
  async def test_div(self):
    text = await render_element(El.div(content=["Hello World!"]))
    self.assertEqual(text, "<div>Hello World!</div>")

  async def test_input(self):
    text = await render_element(VEl.input(type="text"))
    self.assertEqual(text, "<input type=\"text\">")

  async def test_fragment(self):
    text = await render_element(HTMLFragment([
      El.div(content=["Hello"]),
      El.div(content=["World"])
    ]))
    self.assertEqual(text, "<div>Hello</div><div>World</div>")

  async def test_class_map(self):
    text = await render_element(VEl.input(_class=class_map({ "text-input": True })))
    self.assertEqual(text, "<input class=\"text-input\">")

    text = await render_element(VEl.input(_class=class_map({ "text-input": False })))
    self.assertEqual(text, "<input class=\"\">")

  async def test_component(self):
    class TestComp(Component):
      def render(self):
        return El.div(content=["Hello World!"])

    text = await render_element(TestComp())
    self.assertEqual(text, "<div>Hello World!</div>")

if __name__ == "__main__":
  _ = unittest.main()



import codecs
import functools
from io import BytesIO
import json
import mimetypes
import pathlib
from typing import Any, Callable
from collections.abc import  Awaitable, Iterable, MutableMapping

BytesLike = bytes | bytearray
ASGIHeaders = Iterable[tuple[BytesLike, BytesLike]]

ASGIScope = MutableMapping[str, Any]
ASGIFnSend = Callable[[MutableMapping[str, Any]], Awaitable[Any]]
ASGIFnReceive = Callable[[], Awaitable[MutableMapping[str, Any]]]
ASGIHandler = Callable[[ASGIScope, ASGIFnReceive, ASGIFnSend], Awaitable[Any]]

class TransportContext:
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    self._scope = scope
    self._receive = receive
    self._send = send

  @property
  def path(self): return self._scope["path"]
  @property
  def query_string(self) -> str | None: return None if self._scope["query_string"] is None else self._scope["query_string"].decode("utf-8")
  @property
  def fullpath(self): return (self._scope["raw_path"] or b"").decode("utf-8").split("?", 1)[0]
  @property
  def scope(self) -> ASGIScope: return { **self._scope }
  @functools.cached_property
  def headers(self):
    res: dict[str, tuple[str, ...]] = {}
    for k, v in self._scope["headers"]:
      key = k.decode(errors="ignore").lower()
      res[key] = res.get(key, ()) + (v.decode(errors="ignore"),)
    return res
  @functools.cached_property
  def content_type(self):
    ct = self.headers.get("content-type")
    if ct is None or len(ct) == 0: raise ValueError("No content type specified on request!")
    if len(ct) > 1: raise ValueError("More than one content-type was specified!")
    ct = ct[0]
    parts = [ p.strip() for p in ct.split(";") ]
    mime_type = parts[0].lower()
    params = { k.lower(): v for k, v in (tuple(p.split("=") for p in parts[1:] if p.count("=") == 1)) }
    return mime_type, params

  @property
  def location(self):
    location = self.path
    if self.query_string is not None: location += f"?{self.query_string}"
    return location

class WebsocketContext(TransportContext):
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    super().__init__(scope, receive, send)
    self._connected = False
    self._accepted = False
    self._scope: ASGIScope

  @property
  def connected(self): return self._connected

  @property
  def accepted(self): return self._accepted

  async def setup(self, headers: ASGIHeaders = (), subprotocol: str | None = None):
    if self._connected: raise ConnectionError("Already connected!")
    if self._accepted: raise ConnectionError("Already accepted!")

    event = await self._receive()
    if event["type"] != "websocket.connect": raise ConnectionError("Did not receive connect event!")
    await self._send({ "type": "websocket.accept", "subprotocol": subprotocol, "headers": [ (name.lower(), value) for name, value in headers ] })
    self._connected = self._accepted = True

  async def receive_message(self) -> BytesLike | str:
    if not self._accepted: raise ConnectionError("not accepted!")
    if not self._connected: raise ConnectionError("Not connected!")
    while self._connected:
      event = await self._receive()
      if event["type"] == "websocket.disconnect":
        self._connected = False
        raise ConnectionError("Connection closed!")
      elif event["type"] == "websocket.receive":
        return event.get("bytes", event.get("text"))
    raise ConnectionError("Connection closed!")

  async def send_message(self, data: str | BytesLike):
    if not self._accepted: raise ConnectionError("not accepted!")
    if not self._connected: raise ConnectionError("Not connected!")

    event: dict[str, Any] = { "type": "websocket.send", "bytes": None, "text": None }
    if isinstance(data, str): event["text"] = data
    else: event["bytes"] = data
    await self._send(event)

  async def close(self, code: int = 1000, reason: str = "Normal Closure"):
    if not self._accepted: raise ConnectionError("not accepted!")
    if not self._connected: raise ConnectionError("Not connected!")

    await self._send({ "type": "websocket.close", "code": code, "reason": reason })
    self._connected = False

def content_headers(content_length: int, mime_type: str, charset: str | None = None):
  content_type = mime_type
  if charset is not None: content_type += f"; charset={charset}"
  return [
    (b"content-length", str(content_length).encode("utf-8")),
    (b"content-type", content_type.encode("utf-8"))
  ]

class HTTPContext(TransportContext):
  def __init__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> None:
    super().__init__(scope, receive, send)
    self._add_response_headers: list[tuple[BytesLike, BytesLike]] = []
    self._scope: ASGIScope

  @property
  def method(self): return self._scope["method"]

  def add_response_headers(self, headers: ASGIHeaders): self._add_response_headers.extend(headers)

  async def respond_text(self, text: str, status: int = 200, mime_type: str = "text/plain"):
    data = text.encode("utf-8")
    await self.respond_bytes(status, content_headers(len(data), mime_type, "utf-8"), data)

  async def respond_file(self, path: str | pathlib.Path, status: int = 200, mime_type: str | None = None):
    mime_type = mime_type or mimetypes.guess_type(path)[0]
    if mime_type is None: raise ValueError("Unknown mime type!")

    with open(path, "rb") as fd:
      data = fd.read()
      await self.respond_bytes(status, content_headers(len(data), mime_type), data)

  async def respond_bytes(self, status: int, headers: Iterable[tuple[BytesLike, BytesLike]], data: bytes):
    await self._send({
      "type": "http.response.start",
      "status": status,
      "headers": list(headers) + self._add_response_headers,
      "trailers": False
    })
    await self._send({
      "type": "http.response.body",
      "body": data,
      "more_body": False
    })

  async def receive_json(self): return json.loads(await self.receive_json_raw())
  async def receive_json_raw(self): return await self.receive_text({ "application/json" })
  async def receive_text(self, allowed_mime_types: Iterable[str]):
    allowed_mime_types = allowed_mime_types if isinstance(allowed_mime_types, set) else set(allowed_mime_types)
    mime_type, ct_params = self.content_type
    if mime_type not in allowed_mime_types: raise ValueError(f"Mime type '{mime_type}' is not in allowed types!")
    charset = ct_params.get("charset", "utf-8")
    try: decoder = codecs.getdecoder(charset)
    except LookupError: raise ValueError("Invalid content-type encoding!")
    data = await self.receive_data()
    return decoder(data, "ignore")[0]
  async def receive_data(self) -> bytes:
    more = True
    io = BytesIO()
    while more:
      event = await self._receive()
      event_type = event.get("type")
      if event_type == "http.request":
        more = event.get("more_body", False)
        _ = io.write(event.get("body", b""))
      elif event_type == "http.disconnect": more = False
    return io.getvalue()



import unittest
import jwt
from datetime import datetime, timedelta, timezone
from rxxxt.state import JWTStateResolver, StateResolverError

def encode_own_jwt(payload: dict[str, str], secret: str, max_age: timedelta, algorithm: str):
  r = JWTStateResolver(secret.encode("utf-8"), max_age, algorithm)
  return r.create_token(payload, None)

def decode_own_jwt(token: str, secret: str, max_age: timedelta, algorithm: str):
  r = JWTStateResolver(secret.encode("utf-8"), max_age, algorithm)
  return r.resolve(token)

def encode_ref_jwt(payload: dict[str, str], secret: str, max_age: timedelta, algorithm: str):
  data = { "data": payload, "exp": int((datetime.now(tz=timezone.utc) + max_age).timestamp()) }
  return jwt.encode(data, secret, algorithm)

def decode_ref_jwt(token: str, secret: str, _max_age: timedelta, algorithm: str):
  data = jwt.decode(token, secret, [algorithm])
  return data["data"]

class TestJWT(unittest.TestCase):
  def test_encode(self):
    args = ("12345678", timedelta(days=1), "HS512")

    data = { "Hello": "World" }
    token = encode_own_jwt(data, *args)
    token_data = decode_ref_jwt(token, *args)

    self.assertDictEqual(data, token_data)

  def test_decode(self):
    args = ("12345678", timedelta(days=1), "HS512")

    data = { "Hello": "World" }
    token = encode_ref_jwt(data, *args)
    token_data = decode_own_jwt(token, *args)

    self.assertDictEqual(data, token_data)

  def test_expired(self):
    args = ("12345678", timedelta(days=-1), "HS512")
    token = encode_ref_jwt({ "Hello": "World" }, *args)
    with self.assertRaises(StateResolverError):
      _ = decode_own_jwt(token, *args)

  def test_invalid(self):
    args = (timedelta(days=1), "HS512")
    token = encode_ref_jwt({ "Hello": "World" }, "1234", *args)

    _ = decode_own_jwt(token, "1234", *args)
    with self.assertRaises(StateResolverError):
      _ = decode_own_jwt(token, "1235", *args)

if __name__ == "__main__":
  _ = unittest.main()



from rxxxt.elements import HTMLFragment, Element, HTMLElement, HTMLVoidElement, UnescapedHTMLElement, CustomAttribute, VEl, El
from rxxxt.component import Component, event_handler, HandleNavigate
from rxxxt.page import default_page, PageBuilder
from rxxxt.execution import Context
from rxxxt.app import App
from rxxxt.router import Router, router_params
from rxxxt.state import local_state, global_state, context_state, local_state_box, global_state_box, context_state_box
from rxxxt.utils import class_map

__all__ = [
  "HTMLFragment", "Element", "HTMLElement", "HTMLVoidElement", "UnescapedHTMLElement", "CustomAttribute", "VEl", "El",
  "Component", "event_handler", "HandleNavigate",
  "PageBuilder", "default_page",
  "Context",
  "App",
  "Router", "router_params",
  "local_state", "global_state", "context_state", "local_state_box", "global_state_box", "context_state_box",
  "class_map"
]



from abc import ABC, abstractmethod
import base64
from datetime import datetime, timedelta, timezone
import hashlib
import inspect
from io import BytesIO
import json
import os
import secrets
from typing import Awaitable, Callable, Generic, TypeVar, get_origin
from pydantic import TypeAdapter, ValidationError
import hmac

from rxxxt.cell import SerilializableStateCell
from rxxxt.component import Component
from rxxxt.execution import Context, State

T = TypeVar("T")
StateDataAdapter = TypeAdapter(dict[str, str])

class StateBox(Generic[T]):
  def __init__(self, key: str, state: State, inner: SerilializableStateCell[T]) -> None:
    super().__init__()
    self._key = key
    self._state = state
    self._inner = inner

  def __enter__(self): return self
  def __exit__(self, *_): self.update()

  @property
  def key(self): return self._key

  @property
  def value(self): return self._inner.value

  @value.setter
  def value(self, v: T):
    self._inner.value = v
    self.update()

  def update(self): self._state.request_key_updates({ self._key })

class StateBoxDescriptorBase(Generic[T]):
  def __init__(self, state_key_producer: Callable[[Context, str], str], default_factory: Callable[[], T], state_name: str | None = None) -> None:
    self._state_name = state_name
    self._default_factory = default_factory
    self._state_key_producer = state_key_producer

    native_types = (bool, bytearray, bytes, complex, dict, float, frozenset, int, list, object, set, str, tuple)
    if default_factory in native_types or get_origin(default_factory) in native_types:
      self._val_type_adapter = TypeAdapter(default_factory)
    else:
      sig = inspect.signature(default_factory)
      self._val_type_adapter = TypeAdapter(sig.return_annotation)

  def __set_name__(self, owner, name):
    if self._state_name is None:
      self._state_name = name

  def _get_box(self, context: Context) -> StateBox[T]:
    if not self._state_name: raise ValueError("State name not defined!")
    key = self._state_key_producer(context, self._state_name)

    if (cell:=context.state.get_key_cell(key)) is not None:
      if not isinstance(cell, SerilializableStateCell):
        raise ValueError(f"Cell is not serializable for key '{key}'!")
      return StateBox(key, context.state, cell)

    old_str_value = context.state.get_key_str(key)
    if old_str_value is None:
      value = self._default_factory()
    else:
      value = self._val_type_adapter.validate_json(old_str_value)
    cell = SerilializableStateCell(value, lambda v: self._val_type_adapter.dump_json(v).decode("utf-8"))
    context.state.set_key_cell(key, cell)
    return StateBox(key, context.state, cell)

class StateBoxDescriptor(StateBoxDescriptorBase[T]):
  def __get__(self, obj, objtype=None):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    obj.context.subscribe(box.key)

    return box

class StateDescriptor(StateBoxDescriptorBase[T]):
  def __set__(self, obj, value):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    box.value = value

  def __get__(self, obj, objtype=None):
    if not isinstance(obj, Component):
      raise TypeError("StateDescriptor used on non-component!")

    box = self._get_box(obj.context)
    obj.context.subscribe(box.key)

    return box.value

def get_global_state_key(context: Context, name: str):
  return f"global;{name}"

def get_local_state_key(context: Context, name: str):
  return f"#local;{context.sid};{name}"

def get_context_state_key(context: Context, name: str):
  state_key = None
  exisiting_keys = context.state.keys
  for sid in context.stack_sids:
    state_key = f"#context;{sid};{name}"
    if state_key in exisiting_keys:
      return state_key
  if state_key is None: raise ValueError(f"State key not found for context '{name}'!")
  return state_key # this is just the key for context.sid

def local_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_local_state_key, default_factory, state_name=name)

def global_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_global_state_key, default_factory, state_name=name)

def context_state(default_factory: Callable[[], T], name: str | None = None):
  return StateDescriptor(get_context_state_key, default_factory, state_name=name)

def local_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_local_state_key, default_factory, state_name=name)

def global_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_global_state_key, default_factory, state_name=name)

def context_state_box(default_factory: Callable[[], T], name: str | None = None):
  return StateBoxDescriptor(get_context_state_key, default_factory, state_name=name)

class StateResolverError(BaseException): pass

class StateResolver(ABC):
  @abstractmethod
  def create_token(self, data: dict[str, str], old_token: str | None) -> str | Awaitable[str]: pass
  @abstractmethod
  def resolve(self, token: str) -> dict[str, str] | Awaitable[dict[str, str]]: pass

class JWTStateResolver(StateResolver):
  def __init__(self, secret: bytes, max_age: timedelta | None = None, algorithm: str = "HS512") -> None:
    super().__init__()
    self._secret = secret
    self._algorithm = algorithm
    self._digest = { "HS256": hashlib.sha256, "HS384": hashlib.sha384, "HS512": hashlib.sha512 }[algorithm]
    self._max_age: timedelta = timedelta(days=1) if max_age is None else max_age

  def create_token(self, data: dict[str, str], old_token: str | None) -> str:
    payload = { "exp": int((datetime.now(tz=timezone.utc) + self._max_age).timestamp()), "data": data }
    stream = BytesIO()
    stream.write(JWTStateResolver._b64url_encode(json.dumps({
      "typ": "JWT",
      "alg": self._algorithm
    }).encode("utf-8")))
    stream.write(b".")
    stream.write(JWTStateResolver._b64url_encode(json.dumps(payload).encode("utf-8")))

    signature = hmac.digest(self._secret, stream.getvalue(), self._digest)
    stream.write(b".")
    stream.write(JWTStateResolver._b64url_encode(signature))
    return stream.getvalue().decode("utf-8")

  def resolve(self, token: str) -> dict[str, str]:
    rtoken = token.encode("utf-8")
    sig_start = rtoken.rfind(b".")
    if sig_start == -1: raise StateResolverError("Invalid token format")
    parts = rtoken.split(b".")
    if len(parts) != 3: raise StateResolverError("Invalid token format")

    try: header = json.loads(JWTStateResolver._b64url_decode(parts[0]))
    except: raise StateResolverError("Invalid token header")

    if not isinstance(header, dict) or header.get("typ") != "JWT" or header.get("alg") != self._algorithm:
      raise StateResolverError("Invalid header contents")

    signature = JWTStateResolver._b64url_decode(rtoken[(sig_start + 1):])
    actual_signature = hmac.digest(self._secret, rtoken[:sig_start], self._digest)
    if not hmac.compare_digest(signature, actual_signature):
      raise StateResolverError("Invalid JWT signature!")

    payload = json.loads(JWTStateResolver._b64url_decode(parts[1]))
    if not isinstance(payload, dict) or not isinstance(payload.get("exp"), int) or not isinstance(payload.get("data"), dict):
      raise StateResolverError("Invalid JWT payload!")

    expires_dt = datetime.fromtimestamp(payload["exp"], timezone.utc)
    if expires_dt < datetime.now(tz=timezone.utc):
      raise StateResolverError("JWT expired!")

    try: state_data = StateDataAdapter.validate_python(payload["data"])
    except ValidationError as e: raise StateResolverError(e)
    return state_data

  @staticmethod
  def _b64url_encode(value: bytes | bytearray): return base64.urlsafe_b64encode(value).rstrip(b"=")
  @staticmethod
  def _b64url_decode(value: bytes | bytearray): return base64.urlsafe_b64decode(value + b"=" * (4 - len(value) % 4))

def default_state_resolver() -> JWTStateResolver:
  """
  Creates a JWTStateResolver.
  Uses the environment variable `JWT_SECRET` as its secret, if set, otherwise creates a new random, temporary secret.
  """

  jwt_secret = os.getenv("JWT_SECRET", None)
  if jwt_secret is None: jwt_secret = secrets.token_bytes(64)
  else: jwt_secret = jwt_secret.encode("utf-8")
  return JWTStateResolver(jwt_secret)



import asyncio
from dataclasses import dataclass
from pydantic import BaseModel
from rxxxt.elements import El, Element, HTMLFragment, UnescapedHTMLElement, meta_element
from rxxxt.events import InputEvent, OutputEvent
from rxxxt.execution import Context, ContextConfig, State
from rxxxt.helpers import to_awaitable
from rxxxt.page import PageFactory
from rxxxt.renderer import Renderer, render_node
from rxxxt.state import StateResolver

class InitOutputData(BaseModel):
  path: str
  state_token: str
  events: tuple[OutputEvent, ...]

class UpdateOutputData(BaseModel):
  state_token: str | None = None
  events: tuple[OutputEvent, ...]
  html_parts: tuple[str, ...]

@dataclass
class SessionConfig:
  persistent: bool
  page_facotry: PageFactory
  state_resolver: StateResolver

class Session:
  def __init__(self, config: SessionConfig, base: Element) -> None:
    self._update_event = asyncio.Event()
    self.config = config
    self.state = State(self._update_event)

    context_config = ContextConfig(persistent=config.persistent, render_meta=True)
    root_node = meta_element("root", base).tonode(Context(self.state, context_config, ("root",)))
    self._root_renderer = Renderer(root_node)
    self._last_token: str | None = None

  async def __aenter__(self): return self
  async def __aexit__(self, *_): await self.destroy()

  async def wait_for_update(self): _ = await self._update_event.wait()

  async def init(self, state_token: str | None):
    if state_token is not None:
      self._last_token = state_token
      user_data = await to_awaitable(self.config.state_resolver.resolve, state_token)
      self.state.init(user_data)

    await self._root_renderer.expand()

  async def destroy(self):
    await self._root_renderer.destroy()
    self.state.destroy()

  async def update(self):
    await self._root_renderer.update(self.state.pop_updates())
    self.state.cleanup()

  async def handle_events(self, events: tuple[InputEvent, ...]):
    await self._root_renderer.handle_events(events)

  def set_location(self, location: str): self.state.update_state_strs({ "!location": location })
  def set_headers(self, headers: dict[str, tuple[str, ...]]):
    headers_kvs = { f"!header;{k}": "\n".join(v) for k, v in headers.items() }
    olds_header_keys = set(k for k in self.state.keys if k.startswith("!header;"))
    olds_header_keys.difference_update(headers_kvs.keys())
    for k in olds_header_keys: self.state.delete_key(k)
    self.state.update_state_strs(headers_kvs)
    self.state.request_key_updates(olds_header_keys)

  async def render_update(self, include_state_token: bool, render_full: bool):
    state_token: str | None = None
    if include_state_token: state_token = await self._update_state_token()

    html_parts = tuple(self._root_renderer.render_full()) if render_full else self._root_renderer.render_partial()
    return UpdateOutputData(state_token=state_token, html_parts=html_parts, events=self.state.pop_output_events())

  async def render_page(self):
    init_data = InitOutputData(state_token=await self._update_state_token(), events=self.state.pop_output_events(), path="")

    content_el = UnescapedHTMLElement(self._root_renderer.render_full())
    header_el = HTMLFragment([ El.script(src="/rxxxt-client.js"), El.style(content=["rxxxt-meta { display: contents; }"]) ])
    body_end_el = HTMLFragment([ El.script(content=[ UnescapedHTMLElement(f"window.rxxxt.init({init_data.model_dump_json(exclude_defaults=True)});") ]) ])

    page = self.config.page_facotry(header_el, content_el, body_end_el)
    node = page.tonode(Context(self.state, ContextConfig(persistent=False, render_meta=False), ("page",)))
    await node.expand()
    res = render_node(node)
    await node.destroy()
    return res

  async def _update_state_token(self):
    self.state.cleanup()
    self._last_token = await to_awaitable(self.config.state_resolver.create_token, self.state.user_data, self._last_token)
    return self._last_token



from inspect import isawaitable
from typing import Callable, TypeVar, cast, Any
from collections.abc import Awaitable

T = TypeVar("T")
async def to_awaitable(fn: Callable[..., T | Awaitable[T]], *args: Any, **kwargs: Any) -> T:
  result = fn(*args, **kwargs)
  if isawaitable(result): result = await result
  return cast(T, result)



from typing import Protocol
from rxxxt.elements import El, Element, HTMLFragment, VEl
from collections.abc import Iterable

class PageFactory(Protocol):
  def __call__(self, header: Element, content: Element, body_end: Element) -> Element: ...

def default_page(header: Element, content: Element, body_end: Element):
  return HTMLFragment([
    VEl["!DOCTYPE"](html=None),
    El.html(content=[
      El.head(content=[
        VEl.meta(charset="UTF-8"),
        VEl.meta(name="viewport", content="width=device-width, initial-scale=1.0"),
        header
      ]),
      El.body(content=[
        content,
        body_end
      ])
    ])
  ])

class PageBuilder(PageFactory):
  def __init__(self, page_factory: PageFactory = default_page) -> None:
    self._header_elements: list[Element] = []
    self._body_end_elements: list[Element] = []
    self._page_factory = page_factory

  def add_stylesheet(self, url: str, **kwargs: str): self.add_header(VEl.link(rel="stylesheet", href=url, **kwargs))
  def add_header_script(self, url: str, content: Iterable[Element | str] = (), **kwargs: str):
    self.add_header(El.script(src=url, content=content, **kwargs))
  def add_body_script(self, url: str, content: Iterable[Element | str] = (), **kwargs: str):
    self.add_body_end(El.script(src=url, content=content, **kwargs))

  def add_header(self, el: Element): self._header_elements.append(el)
  def add_body_end(self, el: Element): self._body_end_elements.append(el)

  def __call__(self, header: Element, content: Element, body_end: Element) -> Element:
    return self._page_factory(HTMLFragment([ header, *self._header_elements ]), content, HTMLFragment([ *self._body_end_elements, body_end ]))




from abc import ABC
import html
from io import StringIO
from rxxxt.events import InputEvent
from rxxxt.execution import Context

class Node(ABC):
  def __init__(self, context: Context, children: tuple['Node', ...]) -> None:
    self.context = context
    self.children = children

  async def expand(self):
    for c in self.children: await c.expand()

  async def update(self):
    for c in self.children: await c.update()

  async def handle_events(self, events: tuple[InputEvent, ...]):
    for c in self.children: await c.handle_events(events)

  async def destroy(self):
    for c in self.children: await c.destroy()

  def write(self, io: StringIO):
    for c in self.children: c.write(io)

class FragmentNode(Node): ...
class TextNode(Node):
  def __init__(self, context: Context, text: str) -> None:
    super().__init__(context, ())
    self.text = text

  def write(self, io: StringIO): _ = io.write(self.text)

class VoidElementNode(Node):
  def __init__(self, context: Context, tag: str, attributes: dict[str, str | None], children: tuple['Node', ...] = ()) -> None:
    super().__init__(context, children)
    self.attributes = attributes
    self.tag = tag

  def write(self, io: StringIO):
    _ = io.write(f"<{html.escape(self.tag)}")
    for k, v in self.attributes.items():
      _ = io.write(f" {html.escape(k)}")
      if v is not None: _ = io.write(f"=\"{html.escape(v)}\"")
    _ = io.write(">")

class ElementNode(VoidElementNode):
  def __init__(self, context: Context, tag: str, attributes: dict[str, str | None], children: tuple['Node', ...]) -> None:
    super().__init__(context, tag, attributes, children)

  def write(self, io: StringIO):
    super().write(io)
    for c in self.children: c.write(io)
    _ = io.write(f"</{html.escape(self.tag)}>")



import re
from typing import cast
from rxxxt.elements import El, Element, ElementFactory
from rxxxt.component import Component
from rxxxt.state import context_state

class PathPattern:
  def __init__(self, pattern: str) -> None:
    pattern = pattern.rstrip("/") if len(pattern) > 1 else pattern
    param_ranges: list[tuple[int, int]] = []
    param_search_start = 0
    while (param_start := pattern.find("{", param_search_start)) != -1:
      param_end = pattern.find("}", param_start)
      if param_end == -1: raise ValueError("Invalid pattern. Expected closing brace after opening brace.")
      param_search_start = param_end
      param_ranges.append((param_start, param_end))

    param_name_regex = re.compile("^[a-zA-Z0-9_]*\\*?$")

    self.parts: list[str] = []
    self.params: list[tuple[str | None, bool]] = []
    part_start = 0
    for ps, pe in param_ranges:
      self.parts.append(pattern[part_start:ps])

      param_txt = pattern[ps+1:pe].strip()
      if not param_name_regex.match(param_txt): raise ValueError("Invalid parameter name.")
      if param_txt.endswith("*"): self.params.append((param_txt[:-1] or None, True))
      else: self.params.append((param_txt or None, False))

      part_start = pe + 1

    self.parts.append(pattern[part_start:])

    for part in self.parts:
      if "}" in part: raise ValueError("Invalid pattern. Found closing brace without an opening brace.")

  def match(self, path: str) -> dict[str,str] | None:
    # path = path.rstrip("/")
    if len(self.parts) == 1:
      return {} if path == self.parts[0] else None

    params: dict[str, str] = {}
    current_index = 0
    for idx in range(0, len(self.parts) - 1):
      part1 = self.parts[idx]
      part2 = self.parts[idx + 1]
      param_name, param_allow_slash = self.params[idx]

      part1_len = len(part1)
      if path[current_index:current_index + part1_len] != part1: return None

      current_index += part1_len
      if part2 == "" and idx == len(self.parts) - 2:
        param_val = path[current_index:]
      else:
        part2_start = path.find(part2, current_index)
        if part2_start == -1: return None
        param_val = path[current_index:part2_start]
      current_index += len(param_val)
      if "/" in param_val and not param_allow_slash: return None
      if param_name is not None: params[param_name] = param_val
    return params

def router_params(): return context_state(dict[str, str], "*rp*")

class Router(ElementFactory):
  class RoutedComponent(Component):
    params = router_params()

    def __init__(self, routes: list[tuple[PathPattern, ElementFactory]]):
      super().__init__()
      self._routes = routes
      self._selected_match: tuple[ElementFactory, dict[str, str]] | None = None

    async def on_before_update(self) -> None:
      self._selected_match = self._get_current_match()
      self.params = cast(dict[str, str], dict()) if self._selected_match is None else self._selected_match[1]

    def render(self) -> Element:
      if self._selected_match is None:
        return El.h1(content=["Not found!"])
      else:
        return self._selected_match[0]()

    def _get_current_match(self):
      path = self.context.path
      for pattern, element_factory in self._routes:
        if (m:=pattern.match(path)) is not None:
          return (element_factory, m)
      return None


  def __init__(self) -> None:
    self._routes: list[tuple[PathPattern, ElementFactory]] = []

  def add_route(self, path: str, element_factory: ElementFactory): self._routes.append((PathPattern(path), element_factory))
  def route(self, path: str):
    def _inner(fn: ElementFactory):
      self.add_route(path, fn)
      return fn
    return _inner

  def __call__(self) -> Element: return Router.RoutedComponent(self._routes)



from abc import abstractmethod
import asyncio
import base64
import inspect
from typing import Annotated, Any, Callable, Generic, ParamSpec, TypeVar, get_args, get_origin, cast
from collections.abc import Awaitable, Coroutine
from pydantic import validate_call
from rxxxt.elements import CustomAttribute, Element, meta_element
from rxxxt.events import ContextInputEventDescriptor, ContextInputEventDescriptorGenerator, ContextInputEventHandlerOptions, InputEvent
from rxxxt.execution import Context
from rxxxt.helpers import to_awaitable
from rxxxt.node import Node

EHP = ParamSpec('EHP')
EHR = TypeVar('EHR')

class ClassEventHandler(Generic[EHP, EHR]):
  def __init__(self, fn: Callable[EHP, EHR], options: ContextInputEventHandlerOptions) -> None:
    self.fn = fn
    self.options = options
  def __get__(self, instance: Any, _): return InstanceEventHandler(self.fn, self.options, instance)
  def __call__(self, *args: EHP.args, **kwargs: EHP.kwargs) -> EHR: raise RuntimeError("The event handler can only be called when attached to an instance!")

class InstanceEventHandler(ClassEventHandler[EHP, EHR], Generic[EHP, EHR], CustomAttribute, ContextInputEventDescriptorGenerator):
  def __init__(self, fn: Callable[EHP, EHR], options: ContextInputEventHandlerOptions, instance: Any) -> None:
    super().__init__(validate_call(fn), options)
    if not isinstance(instance, Component): raise ValueError("The provided instance must be a component!")
    self.instance = instance

  @property
  def descriptor(self):
    return ContextInputEventDescriptor(
      context_id=self.instance.context.sid,
      handler_name=self.fn.__name__,
      param_map=self._get_param_map(),
      options=self.options)

  def __call__(self, *args: EHP.args, **kwargs: EHP.kwargs) -> EHR: return self.fn(self.instance, *args, **kwargs)

  def get_key_value(self, original_key: str):
    if not original_key.startswith("on"): raise ValueError("Event handler must be applied to an attribute starting with 'on'.")
    v = base64.b64encode(self.descriptor.model_dump_json(exclude_defaults=True).encode("utf-8")).decode("utf-8")
    return (f"rxxxt-on-{original_key[2:]}", v)

  def _get_param_map(self):
    param_map: dict[str, str] = {}
    sig = inspect.signature(self.fn)

    for i, (name, param) in enumerate(sig.parameters.items()):
      if i == 0: continue # skip self

      if get_origin(param.annotation) is Annotated:
        args = get_args(param.annotation)
        metadata = args[1:]

        if len(metadata) < 1:
          raise ValueError(f"Parameter '{name}' is missing the second annotation.")

        param_map[name] = metadata[0]
      else:
        raise TypeError(f"Parameter '{name}' must be of type Annotated.")

    return param_map

def event_handler(**kwargs: Any):
  options = ContextInputEventHandlerOptions.model_validate(kwargs)
  def _inner(fn: Callable[EHP, EHR]) -> ClassEventHandler[EHP, EHR]: return ClassEventHandler(fn, options)
  return _inner

class HandleNavigate(CustomAttribute):
  def __init__(self, location: str) -> None:
    super().__init__()
    self.location = location

  def get_key_value(self, original_key: str) -> tuple[str, str]:
    return (original_key, f"window.rxxxt.navigate('{self.location}');")

class Component(Element):
  def __init__(self) -> None:
    super().__init__()
    self.context: Context
    self._worker_tasks: list[asyncio.Task[Any]] = []
    self._job_tasks: list[asyncio.Task[Any]] = []

  @abstractmethod
  def render(self) -> Element | Awaitable[Element]: ...

  def add_job(self, a: Coroutine[Any, Any, Any]):
    """
    Runs a background job until completion. Only runs when the session is persistent.
    args:
      a: Coroutine - the coroutine that should be run
    """
    if self.context.config.persistent:
      self._worker_tasks.append(asyncio.create_task(a))
    else: a.close()
  def add_worker(self, a: Coroutine[Any, Any, Any]):
    """
    Runs a background worker, which may be cancelled at any time. Only runs when the session is persistent.
    args:
      a: Coroutine - the coroutine that should be run
    """
    if self.context.config.persistent:
      self._worker_tasks.append(asyncio.create_task(a))
    else: a.close()

  async def lc_init(self, context: Context) -> None:
    self.context = context
    await self.on_init()

  async def lc_render(self) -> Element:
    await self.on_before_update()
    el = await to_awaitable(self.render)
    await self.on_after_update()
    return el
  async def lc_destroy(self) -> None:
    await self.on_before_destroy()
    if len(self._job_tasks) > 0:
      try: _ = await asyncio.wait(self._job_tasks)
      except asyncio.CancelledError: pass
      self._job_tasks.clear()
    if len(self._worker_tasks) > 0:
      for t in self._worker_tasks: _ = t.cancel()
      try: _ = await asyncio.wait(self._worker_tasks)
      except asyncio.CancelledError: pass
      self._worker_tasks.clear()
    await self.on_after_destroy()

  async def lc_handle_event(self, event: dict[str, int | float | str | bool]):
    handler_name = event.pop("$handler_name", None)
    if isinstance(handler_name, str):
      fn = getattr(self, handler_name, None) # NOTE: this is risky!!
      if isinstance(fn, InstanceEventHandler):
        await to_awaitable(cast(InstanceEventHandler[..., Any], fn), **event)

  async def on_init(self) -> None: ...
  async def on_before_update(self) -> None: ...
  async def on_after_update(self) -> None: ...
  async def on_before_destroy(self) -> None: ...
  async def on_after_destroy(self) -> None: ...

  def tonode(self, context: Context) -> 'Node': return ComponentNode(context, self)

class ComponentNode(Node):
  def __init__(self, context: Context, element: Component) -> None:
    super().__init__(context, ())
    self.element = element

  async def expand(self):
    if len(self.children) > 0:
      raise ValueError("Can not expand already expanded element!")

    await self.element.lc_init(self.context)
    await self._render_inner()

  async def update(self):
    for c in self.children: await c.destroy()
    self.children = ()
    await self._render_inner()

  async def handle_events(self, events: tuple[InputEvent, ...]):
    for e in events:
      if self.context.sid == e.context_id:
        await self.element.lc_handle_event(dict(e.data))
    await super().handle_events(events)

  async def destroy(self):
    for c in self.children: await c.destroy()
    self.children = ()

    await self.element.lc_destroy()
    self.context.unsubscribe_all()

  async def _render_inner(self):
    inner = await self.element.lc_render()
    if self.context.config.render_meta:
      inner = meta_element(self.context.sid, inner)
    self.children = (inner.tonode(self.context.sub("inner")),)
    await self.children[0].expand()



import asyncio
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
import functools
import hashlib
import itertools
from typing import Literal
from pydantic import TypeAdapter
from rxxxt.cell import StateCell, StrStateCell
from rxxxt.events import ContextInputEventDescriptor, EventRegisterQuerySelectorEvent, NavigateOutputEvent, \
  OutputEvent, UseWebsocketOutputEvent, SetCookieOutputEvent, EventRegisterWindowEvent, ContextInputEventDescriptorGenerator

ContextStackKey = str | int
ContextStack = tuple[ContextStackKey, ...]
HeaderValuesAdapter = TypeAdapter(tuple[str, ...])

class State:
  """
  State keys may have prefixes. These prefixes inidcate how and when and if a key should be removed from the state.
  Prefixes:
    "#" = temporary - removed from the user data and session state, if no longer used
    "!" = protocol - removed from user state, if not used but not purged from the session
  """

  def __init__(self, update_event: asyncio.Event) -> None:
    self._key_str_store: dict[str, str] = {}
    self._key_cell_store: dict[str, StateCell] = {}

    self._key_subscribers: dict[str, set[ContextStack]] = {}
    self._pending_updates: set[ContextStack] = set()
    self._output_events: list[OutputEvent] = []
    self._update_event = update_event

  @property
  def user_data(self):
    active_keys = self._get_active_keys({ "!", "#" })
    data = { k: v for k, v in self._key_str_store.items() if k in active_keys }
    data.update({ k: cell.serlialize() for k, cell in self._key_cell_store.items() if k in active_keys })
    return data

  @property
  def keys(self) -> set[str]: return set(itertools.chain(self._key_str_store.keys(), self._key_cell_store.keys()))

  def init(self, k_str_store: dict[str, str]): self._key_str_store.update(k_str_store)

  def get_key_str(self, key: str):
    if (v := self._key_str_store.get(key)) is None:
      cell = self._key_cell_store.get(key)
      if cell is not None: v = cell.serlialize()
    return v
  def get_key_cell(self, key: str): return self._key_cell_store.get(key)
  def set_key_cell(self, key: str, cell: StateCell, overwrite: bool = False):
    if key in self._key_cell_store and not overwrite:
      raise ValueError(f"Cell already registered with key '{key}'!")
    self._key_cell_store[key] = cell
  def delete_key(self, key: str):
    _ = self._key_str_store.pop(key, None)
    if (cell := self._key_cell_store.pop(key, None)) is not None:
      cell.destroy()

  def update_state_strs(self, data: dict[str, str]):
    for k, v in data.items():
      cell = self.get_key_cell(k)
      if cell is None:
        self.set_key_cell(k, StrStateCell(v))
      elif isinstance(cell, StrStateCell):
        cell.value = v
      else:
        raise ValueError("invalid cell for location!")
    self.request_key_updates(set(data.keys()))

  def request_context_updates(self, ids: set[ContextStack]):
    for id in ids: self._pending_updates.add(id)
    self._set_update_event()

  def request_key_updates(self, keys: set[str]):
    ids: set[ContextStack] = set(itertools.chain(*(self._key_subscribers.get(key, ()) for key in keys)))
    self.request_context_updates(ids)

  def subscribe(self, cid: ContextStack, key: str):
    self._key_subscribers.setdefault(key, set()).add(cid)

  def unsubscribe(self, cid: ContextStack, key: str):
    subs = self._key_subscribers.get(key)
    if subs is not None and cid in subs: subs.remove(cid)

  def unsubscribe_all(self, cid: ContextStack):
    if cid in self._pending_updates: self._pending_updates.remove(cid)
    for ids in self._key_subscribers.values():
      if cid in ids: ids.remove(cid)
    self._set_update_event()

  def add_output_event(self, event: OutputEvent):
    self._output_events.append(event)
    self._set_update_event()

  def pop_output_events(self):
    res = tuple(self._output_events)
    self._output_events = []
    return res

  def pop_updates(self):
    res = self._pending_updates
    self._pending_updates = set()
    self._set_update_event()
    return res

  def destroy(self):
    for cell in self._key_cell_store.values():
      cell.destroy()
    self._key_cell_store.clear()

  def cleanup(self):
    active_keys = self._get_active_keys({ "#" })
    self._key_str_store = { k: v for k, v in self._key_str_store.items() if k in active_keys }
    inactive_cells = set(self._key_cell_store.keys())
    inactive_cells.difference_update(active_keys)
    for k in inactive_cells:
      if (cell := self._key_cell_store.pop(k, None)) is not None:
        cell.destroy()
    self._key_subscribers = { k: v for k, v in self._key_subscribers.items() if len(v) > 0 }

  def _get_active_keys(self, inactive_prefixes: set[str]):
    active_keys = set(k for k, v in self._key_subscribers.items() if len(v) > 0)
    active_keys.update(k for k in self._key_str_store.keys() if len(k) == 0 or k[0] not in inactive_prefixes)
    active_keys.update(k for k in self._key_cell_store.keys() if len(k) == 0 or k[0] not in inactive_prefixes)
    return active_keys

  def _set_update_event(self):
    if len(self._pending_updates) > 0 or len(self._output_events) > 0: self._update_event.set()
    else: self._update_event.clear()

@functools.lru_cache(maxsize=256)
def get_context_stack_sid(stack: ContextStack):
  hasher = hashlib.sha256()
  for k in stack:
    if isinstance(k, str): k = k.replace(";", ";;")
    else: k = str(k)
    hasher.update((k + ";").encode("utf-8"))
  return hasher.hexdigest()

@dataclass(frozen=True)
class ContextConfig:
  persistent: bool
  render_meta: bool

class Context:
  def __init__(self, state: State, config: ContextConfig, stack: ContextStack) -> None:
    self._stack: ContextStack = stack
    self._config = config
    self.state = state

  @property
  def config(self): return self._config

  @property
  def id(self): return self._stack

  @cached_property
  def sid(self): return get_context_stack_sid(self._stack)

  @property
  def stack_sids(self):
    return [ get_context_stack_sid(self._stack[:i + 1]) for i in range(len(self._stack)) ]

  @property
  def location(self):
    res = self._get_state_str_subscribe("!location")
    if res is None: raise ValueError("No location!")
    else: return res

  @property
  def path(self): return self.location.split("?")[0]

  @property
  def query_string(self):
    parts = self.location.split("?")
    if len(parts) < 2: return None
    else: return parts[1]

  @property
  def cookies(self) -> dict[str, str]:
    values = self.get_header("cookie")
    if len(values) == 0: return {}
    result: dict[str, str] = {}
    for cookie in values[0].split(";"):
      try:
        eq_idx = cookie.index("=")
        result[cookie[:eq_idx]] = cookie[(eq_idx + 1):]
      except ValueError: pass
    return result

  def sub(self, key: ContextStackKey): return Context(self.state, self._config, self._stack + (key,))
  def replace_index(self, key: str):
    if isinstance(self._stack[-1], int): return Context(self.state, self._config, self._stack[:-1] + (key,))
    raise ValueError("No index to replace!")

  def get_header(self, name: str) -> tuple[str, ...]:
    header_lines = self._get_state_str_subscribe(f"!header;{name}")
    if header_lines is None: return ()
    else: return tuple(header_lines.splitlines())

  def request_update(self): self.state.request_context_updates({ self.id })
  def subscribe(self, key: str): self.state.subscribe(self.id, key)
  def unsubscribe(self, key: str): self.state.unsubscribe(self.id, key)
  def unsubscribe_all(self): self.state.unsubscribe_all(self.id)

  def add_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator):
    self._modify_window_event(name, descriptor, "add")
  def add_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool = False):
    self._modify_query_selector_event(selector, name, descriptor, all, "add")

  def remove_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator):
    self._modify_window_event(name, descriptor, "remove")
  def remove_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool = False):
    self._modify_query_selector_event(selector, name, descriptor, all, "remove")

  def navigate(self, location: str):
    self.state.update_state_strs({"!location": location})
    self.state.add_output_event(NavigateOutputEvent(location=location))
  def use_websocket(self, websocket: bool = True): self.state.add_output_event(UseWebsocketOutputEvent(websocket=websocket))
  def set_cookie(self, name: str, value: str, expires: datetime | None = None, path: str | None = None,
                secure: bool | None = None, http_only: bool | None = None, domain: str | None = None, max_age: int | None = None):
    self.state.add_output_event(SetCookieOutputEvent(name=name, value=value, expires=expires, path=path, secure=secure, http_only=http_only, domain=domain, max_age=max_age))
  def delete_cookie(self, name: str):
    self.state.add_output_event(SetCookieOutputEvent(name=name, max_age=-1))

  def _get_state_str_subscribe(self, key: str):
    res = self.state.get_key_str(key)
    self.state.subscribe(self.id, key)
    return res
  def _modify_window_event(self, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, mode: Literal["add"] | Literal["remove"]):
    descriptor = descriptor.descriptor if isinstance(descriptor, ContextInputEventDescriptorGenerator) else descriptor
    self.state.add_output_event(EventRegisterWindowEvent(name=name, mode=mode, descriptor=descriptor))
  def _modify_query_selector_event(self, selector: str, name: str, descriptor: ContextInputEventDescriptor | ContextInputEventDescriptorGenerator, all: bool, mode: Literal["add"] | Literal["remove"]):
    descriptor = descriptor.descriptor if isinstance(descriptor, ContextInputEventDescriptorGenerator) else descriptor
    self.state.add_output_event(EventRegisterQuerySelectorEvent(
      name=name,
      mode=mode,
      selector=selector,
      all=all,
      descriptor=descriptor))



import importlib.resources
import os
from typing import Annotated
import unittest
from rxxxt.app import App, AppHttpRequest, AppWebsocketInitMessage, AppWebsocketUpdateMessage
from rxxxt.asgi import ASGIHandler
from rxxxt.component import Component, event_handler
from rxxxt.elements import El, Element
import httpx
import httpx_ws
import httpx_ws.transport

from rxxxt.events import ContextInputEvent
from rxxxt.state import default_state_resolver, local_state

class TestApp(unittest.IsolatedAsyncioTestCase):

  def _get_client(self, app: ASGIHandler, websocket: bool = False):
    if websocket: transport = httpx_ws.transport.ASGIWebSocketTransport(app)
    else: transport = httpx.ASGITransport(app=app)
    return httpx.AsyncClient(transport=transport, base_url="http://testserver")

  async def test_value_error(self):
    def el_fac() -> Element:
      raise ValueError("Testing...")

    app = App(el_fac)
    async with self._get_client(app) as client:
      r = await client.get("/")
      self.assertEqual(r.status_code, 400)

  async def test_generic_error(self):
    def el_fac() -> Element:
      raise Exception("Testing...")

    app = App(el_fac)
    async with self._get_client(app) as client:
      r = await client.get("/")
      self.assertEqual(r.status_code, 500)

  async def test_not_found_error(self):
    app = App(lambda: El.div())
    async with self._get_client(app) as client:
      r = await client.put("/")
      self.assertEqual(r.status_code, 404)

  async def test_post(self):
    state_resolver = default_state_resolver()
    app = App(lambda: El.div(), state_resolver)
    async with self._get_client(app) as client:
      token = state_resolver.create_token({}, None)
      r = await client.post("/", json=AppHttpRequest(state_token=token, events=()).model_dump())
      self.assertEqual(r.status_code, 200)

  async def test_basic(self):
    text = "This is a test of the app!"
    app = App(lambda: El.div(content=[text]))
    async with self._get_client(app) as client:
      r = await client.get("/")
      self.assertIn(text, r.text)

  @unittest.skipIf(int(os.getenv("CI", "0")), "skipped in CI, build not present there.")
  async def test_frontend_script(self):
    app = App(lambda: El.div())
    async with self._get_client(app) as client:
      r = await client.get("/rxxxt-client.js")
      ref_text = importlib.resources.read_text("rxxxt.assets", "main.js")
      self.assertEqual(r.text, ref_text)

  async def test_ws(self):
    context_id: str = ""
    class Counter(Component):
      counter = local_state(int)

      async def on_init(self) -> None:
        nonlocal context_id
        context_id = self.context.sid

      @event_handler()
      def add(self, value: Annotated[int, "target.value"]):
        self.counter += value

      def render(self):
        return El.div(content=[f"c{self.counter}"])

    state_resolver = default_state_resolver()
    app = App(Counter, state_resolver)
    client = self._get_client(app, True)
    _ = await client.get("/")
    token = state_resolver.create_token({}, None)
    async with httpx_ws.aconnect_ws(str(client.base_url), client) as ws:
      await ws.send_text(AppWebsocketInitMessage(type="init", state_token=token, enable_state_updates=False).model_dump_json())
      await ws.send_text(AppWebsocketUpdateMessage(type="update", events=(
        ContextInputEvent(context_id=context_id, data={ "$handler_name": "add", "value": 5 }),), location="/").model_dump_json())
      response_text = await ws.receive_text()
      self.assertIn("c5", response_text)


if __name__ == "__main__":
  _ = unittest.main()




from io import StringIO
from rxxxt.events import InputEvent
from rxxxt.execution import ContextStack
from rxxxt.node import Node

def render_node(node: Node) -> str:
  io = StringIO()
  node.write(io)
  return io.getvalue()

class Renderer:
  def __init__(self, root: Node) -> None:
    self._root = root
    self._pending_renders: set[ContextStack] = set()

  async def expand(self): await self._root.expand()

  async def update(self, ids: set[ContextStack]):
    for node in self._find_roots(ids):
      self._pending_renders.add(node.context.id)
      await node.update()

  async def handle_events(self, events: tuple[InputEvent, ...]):
    await self._root.handle_events(events)

  async def destroy(self): await self._root.destroy()

  def render_full(self) -> str:
    self._pending_renders.clear()
    return render_node(self._root)

  def render_partial(self):
    res = tuple(render_node(node) for node in self._find_roots(self._pending_renders))
    self._pending_renders.clear()
    return res

  def _find_roots(self, ids: set[ContextStack]):
    if self._root.context.id in ids:
      yield self._root
      return
    els: list[Node] = [self._root]

    while len(els) > 0:
      nels: list[Node] = []
      for nel in (nel for el in els for nel in el.children):
        if nel.context.id in ids: yield nel
        else: nels.append(nel)
      els = nels



from abc import ABC, abstractmethod
import html
import logging
from typing import Protocol
from collections.abc import Iterable
from rxxxt.execution import Context
from rxxxt.node import ElementNode, FragmentNode, Node, TextNode, VoidElementNode

class CustomAttribute(ABC):
  @abstractmethod
  def get_key_value(self, original_key: str) -> tuple[str, str | None]: ...

class Element(ABC):
  @abstractmethod
  def tonode(self, context: Context) -> Node: ...

ElementContent = Iterable[Element | str]
HTMLAttributeValue = str | bool | int | float | CustomAttribute | None

def element_content_to_nodes(context: Context, content: ElementContent):
  nodes: list[Node] = []
  for idx, c in enumerate(content):
    scontext = context.sub(idx)
    if isinstance(c, Element): nodes.append(c.tonode(scontext))
    else: nodes.append(TextNode(scontext, html.escape(c)))
  return tuple(nodes)

class HTMLFragment(Element):
  def __init__(self, content: ElementContent) -> None:
    super().__init__()
    self._content = tuple(content)

  def tonode(self, context: Context) -> Node:
    return FragmentNode(context, element_content_to_nodes(context, self._content))

class HTMLVoidElement(Element):
  def __init__(self, tag: str, attributes: dict[str, HTMLAttributeValue]) -> None:
    super().__init__()
    self._tag = tag
    self._attributes: dict[str, str | None] = {}
    for k, v in attributes.items():
      if isinstance(v, CustomAttribute): k, v = v.get_key_value(k)
      elif isinstance(v, bool):
        if not v: continue
        v = None
      elif isinstance(v, (int, float)): v = str(v)
      self._attributes[k] = v

  def tonode(self, context: Context) -> 'Node':
    return VoidElementNode(context, self._tag, self._attributes)

class HTMLElement(HTMLVoidElement):
  def __init__(self, tag: str, attributes: dict[str, HTMLAttributeValue], content: ElementContent) -> None:
    super().__init__(tag, attributes)
    self._content = tuple(content)

  def tonode(self, context: Context) -> 'Node':
    return ElementNode(context, self._tag, self._attributes, element_content_to_nodes(context, self._content))

class KeyedElement(Element):
  def __init__(self, key: str, element: Element) -> None:
    super().__init__()
    self._key = key
    self._element = element

  def tonode(self, context: Context) -> 'Node':
    try: context = context.replace_index(self._key)
    except ValueError as e: logging.debug(f"Failed to replace index with key {self._key}", e)
    return self._element.tonode(context)

class UnescapedHTMLElement(Element):
  def __init__(self, text: str) -> None:
    super().__init__()
    self._text = text

  def tonode(self, context: Context) -> 'Node': return TextNode(context, self._text)

class CreateHTMLElement(Protocol):
  def __call__(self, content: Iterable[Element | str] = (), key: str | None = None, **kwargs: HTMLAttributeValue) -> Element: ...

class _El(type):
  def __getitem__(cls, name: str) -> CreateHTMLElement:
    def _inner(content: Iterable[Element | str] = (), key: str | None = None, **kwargs: HTMLAttributeValue):
      el = HTMLElement(name, attributes={ k.lstrip("_"): v for k,v in kwargs.items() }, content=list(content))
      if key is not None: el = KeyedElement(key, el)
      return el
    return _inner
  def __getattribute__(cls, name: str): return cls[name]

class El(metaclass=_El): ...

class CreateHTMLVoidElement(Protocol):
  def __call__(self, **kwargs: HTMLAttributeValue) -> HTMLVoidElement: ...

class _VEl(type):
  def __getitem__(cls, name: str) -> CreateHTMLVoidElement:
    def _inner(**kwargs: HTMLAttributeValue) -> HTMLVoidElement:
      return HTMLVoidElement(name, attributes={ k.lstrip("_"): v for k,v in kwargs.items() })
    return _inner
  def __getattribute__(cls, name: str): return cls[name]

class VEl(metaclass=_VEl): ...

class ElementFactory(Protocol):
  def __call__(self) -> Element: ...

def meta_element(id: str, inner: Element): return HTMLElement("rxxxt-meta", {"id":id}, [inner])



import unittest
from typing import Annotated
from rxxxt.component import Component, event_handler
from rxxxt.elements import El
from rxxxt.events import ContextInputEvent
from rxxxt.state import local_state
from tests.helpers import element_to_node, render_node


class TestComponents(unittest.IsolatedAsyncioTestCase):
  class TestCompCounter(Component):
    counter = local_state(int)

    @event_handler()
    def add(self, value: Annotated[int, "target.value"]):
      self.counter += value

    def render(self):
      return El.div(content=[f"c{self.counter}"])

  async def test_component(self):
    comp = TestComponents.TestCompCounter()
    node = element_to_node(comp)
    await node.expand()
    self.assertEqual(render_node(node), "<div>c0</div>")
    comp.counter = 1
    await node.update()
    self.assertEqual(render_node(node), "<div>c1</div>")
    await node.destroy()

  async def test_event_add(self):
     comp = TestComponents.TestCompCounter()
     node = element_to_node(comp)
     await node.expand()
     self.assertEqual(render_node(node), "<div>c0</div>")

     await node.handle_events((ContextInputEvent(context_id=node.context.sid, data={ "$handler_name": "add", "value": 5 }),))
     await node.update()
     self.assertEqual(render_node(node), "<div>c5</div>")
     await node.destroy()

  async def test_double_expand(self):
    el = TestComponents.TestCompCounter()
    node = element_to_node(el)
    await node.expand()
    with self.assertRaises(Exception):
      await node.expand()

if __name__ == "__main__":
  _ = unittest.main()




from abc import ABC, abstractmethod
from datetime import datetime
import re
from typing import Any, Callable, Literal
from pydantic import BaseModel, field_serializer, field_validator, model_serializer

class ContextInputEventHandlerOptions(BaseModel):
  debounce: int | None = None
  throttle: int | None = None
  prevent_default: bool = False

class ContextInputEventDescriptor(BaseModel):
  context_id: str
  handler_name: str
  param_map: dict[str, str]
  options: ContextInputEventHandlerOptions

class ContextInputEventDescriptorGenerator(ABC): # NOTE: this is a typing hack
  @property
  @abstractmethod
  def descriptor(self) -> ContextInputEventDescriptor: pass


class EventBase(BaseModel):
  @model_serializer(mode="wrap")
  def serialize_model(self, old_serizalizer: Callable[[Any], Any]): # TODO: fix typing...
    other = old_serizalizer(self)
    event_name = getattr(self, "event", None)
    if event_name is not None: other["event"] = event_name
    return other

class EventRegisterWindowEvent(EventBase):
  event: Literal["event-modify-window"] = "event-modify-window"
  mode: Literal["add"] | Literal["remove"]
  name: str
  descriptor: ContextInputEventDescriptor

class EventRegisterQuerySelectorEvent(EventBase):
  event: Literal["event-modify-query-selector"] = "event-modify-query-selector"
  mode: Literal["add"] | Literal["remove"]
  name: str
  selector: str
  all: bool
  descriptor: ContextInputEventDescriptor

class SetCookieOutputEvent(EventBase):
  event: Literal["set-cookie"] = "set-cookie"
  name: str
  value: str | None = None
  expires: datetime | None = None
  path: str | None = None
  max_age: int | None = None
  secure: bool | None = None
  http_only: bool | None = None
  domain: str | None = None

  @field_validator('name')
  @classmethod
  def validate_name(cls, value: str):
    if not re.match(r'^[^=;, \t\n\r\f\v]+$', value): raise ValueError("Invalid cookie name")
    return value

  @field_validator('value', "domain")
  @classmethod
  def validate_value(cls, value: str | None):
    if value is None: return None
    if not re.match(r'^[^;, \t\n\r\f\v]+$', value): raise ValueError("Invalid value.")
    return value

  @field_validator('path')
  @classmethod
  def validate_path(cls, value: str | None):
    if value is None: return None
    if not re.match(r'^[^\x00-\x20;,\s]+$', value): raise ValueError("Invalid path value")
    return value

  @field_serializer('expires', when_used='json')
  def seriliaze_expires(self, value: datetime | None): return None if value is None else value.isoformat()

  def to_set_cookie_header(self):
    parts: list[str] = [f"{self.name}={self.value}"]
    if self.path is not None: parts.append(f"path={self.path}")
    if self.expires is not None: parts.append(f"expires={self.expires.strftime('%a, %d %b %G %T %Z')}")
    if self.max_age is not None: parts.append(f"max-age={self.max_age}")
    if self.domain is not None: parts.append(f"domain={self.domain}")
    if self.secure: parts.append("secure")
    if self.http_only: parts.append("httponly")
    return ";".join(parts)

class UseWebsocketOutputEvent(EventBase):
  event: Literal["use-websocket"] = "use-websocket"
  websocket: bool

class NavigateOutputEvent(EventBase):
  event: Literal["navigate"] = "navigate"
  location: str

class ContextInputEvent(BaseModel):
  context_id: str
  data: dict[str, int | float | str | bool | None]

InputEvent = ContextInputEvent
OutputEvent = SetCookieOutputEvent | NavigateOutputEvent | UseWebsocketOutputEvent | EventRegisterWindowEvent | EventRegisterQuerySelectorEvent



import unittest
from rxxxt.elements import El
from rxxxt.router import Router
from tests.helpers import element_to_node, render_node

class TestRouter(unittest.IsolatedAsyncioTestCase):
  async def test_basic(self):
    router = Router()
    router.add_route("/hello", lambda: El.div(content=["hello"]))
    router.add_route("/world", lambda: El.div(content=["world"]))

    comp = router()
    node = element_to_node(comp)
    node.context.state.update_state_strs({ "!location": "/hello" })
    await node.expand()
    self.assertEqual(render_node(node), "<div>hello</div>")

    node.context.state.update_state_strs({ "!location": "/world" })
    await node.update()
    self.assertEqual(render_node(node), "<div>world</div>")

    node.context.state.update_state_strs({ "!location": "/no" })
    await node.update()
    self.assertEqual(render_node(node), "<h1>Not found!</h1>")

    await node.destroy()

  async def test_var_path(self):
    router = Router()
    router.add_route("/var/{value}", lambda: El.div(content=["var1"]))
    router.add_route("/var/{a}/{b}", lambda: El.div(content=["var2"]))

    @router.route("/{path*}")
    def _(): return El.div(content=["not found"])

    comp = router()
    node = element_to_node(comp)
    node.context.state.update_state_strs({ "!location": "/hello" })
    await node.expand()
    self.assertEqual(render_node(node), "<div>not found</div>")

    node.context.state.update_state_strs({ "!location": "/var/1" })
    await node.update()
    self.assertEqual(render_node(node), "<div>var1</div>")

    node.context.state.update_state_strs({ "!location": "/var/1/2" })
    await node.update()
    self.assertEqual(render_node(node), "<div>var2</div>")

    await node.destroy()

if __name__ == "__main__":
  _ = unittest.main()



import asyncio
import contextlib
import importlib.resources
import logging
from typing import Any, Literal
from pydantic import BaseModel, ValidationError
from rxxxt.asgi import ASGIFnReceive, ASGIFnSend, ASGIScope, HTTPContext, WebsocketContext
from rxxxt.elements import ElementFactory
from rxxxt.events import InputEvent
from rxxxt.page import PageFactory, default_page
from rxxxt.session import Session, SessionConfig
from rxxxt.state import StateResolver, default_state_resolver

class AppHttpRequest(BaseModel):
  state_token: str
  events: tuple[InputEvent, ...]

class AppWebsocketInitMessage(BaseModel):
  type: Literal["init"]
  state_token: str
  enable_state_updates: bool

class AppWebsocketUpdateMessage(BaseModel):
  type: Literal["update"]
  events: tuple[InputEvent, ...]
  location: str

class App:
  def __init__(self, content: ElementFactory, state_resolver: StateResolver | None = None, page_factory: PageFactory = default_page) -> None:
    self._content = content
    self._page_factory: PageFactory = page_factory
    self._state_resolver = state_resolver or default_state_resolver()

  async def __call__(self, scope: ASGIScope, receive: ASGIFnReceive, send: ASGIFnSend) -> Any:
    if scope["type"] == "http":
      context = HTTPContext(scope, receive, send)
      try: await self._handle_http(context)
      except asyncio.CancelledError: raise
      except (ValidationError, ValueError) as e:
        logging.debug(e)
        return await context.respond_text("bad request", 400)
      except BaseException as e:
        logging.debug(e)
        return await context.respond_text("internal server error", 500)
    elif scope["type"] == "websocket":
      context = WebsocketContext(scope, receive, send)
      try: await self._ws_session(context)
      except asyncio.CancelledError: raise
      except BaseException as e:
        logging.debug(e)
        await context.close(1011, "Internal error")
      finally:
        if context.connected: await context.close()

  async def _ws_session(self, context: WebsocketContext):
    await context.setup()
    message = await context.receive_message()
    init_message = AppWebsocketInitMessage.model_validate_json(message)

    with contextlib.suppress(ConnectionError):
      async with Session(self._get_session_config(True), self._content()) as session:
        updating_lock = asyncio.Lock()

        async def updater():
          while context.connected:
            await session.wait_for_update()
            async with updating_lock:
              await session.update()
              data = await session.render_update(include_state_token=init_message.enable_state_updates, render_full=False)
              await context.send_message(data.model_dump_json(exclude_defaults=True))

        await session.init(init_message.state_token)

        session.set_location(context.location)
        session.set_headers(context.headers)

        updater_task = asyncio.create_task(updater())
        try:
          while True:
            message = await context.receive_message()
            async with updating_lock:
              update_message = AppWebsocketUpdateMessage.model_validate_json(message)
              session.set_location(update_message.location)
              await session.handle_events(update_message.events)
        finally: _ = updater_task.cancel()

  async def _http_session(self, context: HTTPContext):
    async with Session(self._get_session_config(False), self._content()) as session:

      if context.method == "POST":
        req = AppHttpRequest.model_validate_json(await context.receive_json_raw())
        await session.init(req.state_token)
        events = req.events
      else:
        session.set_location(context.location)
        await session.init(None)
        events = ()

      session.set_location(context.location)
      session.set_headers(context.headers)

      await session.handle_events(events)
      await session.update()

      if context.method == "POST":
        result = await session.render_update(include_state_token=True, render_full=False)
        await context.respond_text(result.model_dump_json(exclude_defaults=True), mime_type="application/json")
      else:
        result = await session.render_page()
        await context.respond_text(result, mime_type="text/html")

  async def _handle_http(self, context: HTTPContext):
    if context.path == "/rxxxt-client.js":
      with importlib.resources.path("rxxxt.assets", "main.js") as file_path:
        await context.respond_file(file_path)
    elif context.method in [ "GET", "POST" ]: await self._http_session(context)
    else: await context.respond_text("not found", 404)

  def _get_session_config(self, persistent: bool):
    return SessionConfig(page_facotry=self._page_factory, state_resolver=self._state_resolver, persistent=persistent)
